@@ -20,9 +20,7 @@
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
-#ifdef HAVE_CONFIG_H
 #include <config.h>
-#endif
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -34,6 +32,7 @@
 #include <glib.h>
 #include <glib-unix.h>
 #include "udscs.h"
+#include "vdagentd-proto-strings.h"
 
 struct udscs_buf {
     uint8_t *buf;
@@ -45,8 +44,6 @@ struct udscs_buf {
 
 struct udscs_connection {
     int fd;
-    const char * const *type_to_string;
-    int no_types;
     int debug;
     void *user_data;
 #ifndef UDSCS_NO_SERVER
@@ -78,27 +75,38 @@ static gboolean udscs_io_channel_cb(GIOChannel *source,
                                     GIOCondition condition,
                                     gpointer data);
 
+static void debug_print_message_header(struct udscs_connection     *conn,
+                                       struct udscs_message_header *header,
+                                       const gchar                 *direction)
+{
+    const gchar *type = "invalid message";
+
+    if (conn == NULL || conn->debug == FALSE)
+        return;
+
+    if (header->type < G_N_ELEMENTS(vdagentd_messages))
+        type = vdagentd_messages[header->type];
+
+    syslog(LOG_DEBUG, "%p %s %s, arg1: %u, arg2: %u, size %u",
+        conn, direction, type, header->arg1, header->arg2, header->size);
+}
+
 struct udscs_connection *udscs_connect(const char *socketname,
     udscs_read_callback read_callback,
     udscs_disconnect_callback disconnect_callback,
-    const char * const type_to_string[], int no_types, int debug)
+    int debug)
 {
     int c;
     struct sockaddr_un address;
     struct udscs_connection *conn;
 
-    conn = calloc(1, sizeof(*conn));
-    if (!conn)
-        return NULL;
-
-    conn->type_to_string = type_to_string;
-    conn->no_types = no_types;
+    conn = g_new0(struct udscs_connection, 1);
     conn->debug = debug;
 
     conn->fd = socket(PF_UNIX, SOCK_STREAM, 0);
     if (conn->fd == -1) {
         syslog(LOG_ERR, "creating unix domain socket: %m");
-        free(conn);
+        g_free(conn);
         return NULL;
     }
 
@@ -109,7 +117,8 @@ struct udscs_connection *udscs_connect(const char *socketname,
         if (conn->debug) {
             syslog(LOG_DEBUG, "connect %s: %m", socketname);
         }
-        free(conn);
+        close(conn->fd);
+        g_free(conn);
         return NULL;
     }
 
@@ -147,13 +156,12 @@ void udscs_destroy_connection(struct udscs_connection **connp)
     wbuf = conn->write_buf;
     while (wbuf) {
         next_wbuf = wbuf->next;
-        free(wbuf->buf);
-        free(wbuf);
+        g_free(wbuf->buf);
+        g_free(wbuf);
         wbuf = next_wbuf;
     }
 
-    free(conn->data.buf);
-    conn->data.buf = NULL;
+    g_clear_pointer(&conn->data.buf, g_free);
 
     if (conn->next)
         conn->next->prev = conn->prev;
@@ -171,8 +179,7 @@ void udscs_destroy_connection(struct udscs_connection **connp)
     if (conn->debug)
         syslog(LOG_DEBUG, "%p disconnected", conn);
 
-    free(conn);
-    *connp = NULL;
+    g_clear_pointer(connp, g_free);
 }
 
 void udscs_set_user_data(struct udscs_connection *conn, void *data)
@@ -188,24 +195,17 @@ void *udscs_get_user_data(struct udscs_connection *conn)
     return conn->user_data;
 }
 
-int udscs_write(struct udscs_connection *conn, uint32_t type, uint32_t arg1,
+void udscs_write(struct udscs_connection *conn, uint32_t type, uint32_t arg1,
     uint32_t arg2, const uint8_t *data, uint32_t size)
 {
     struct udscs_buf *wbuf, *new_wbuf;
     struct udscs_message_header header;
 
-    new_wbuf = malloc(sizeof(*new_wbuf));
-    if (!new_wbuf)
-        return -1;
-
+    new_wbuf = g_new(struct udscs_buf, 1);
     new_wbuf->pos = 0;
     new_wbuf->size = sizeof(header) + size;
     new_wbuf->next = NULL;
-    new_wbuf->buf = malloc(new_wbuf->size);
-    if (!new_wbuf->buf) {
-        free(new_wbuf);
-        return -1;
-    }
+    new_wbuf->buf = g_malloc(new_wbuf->size);
 
     header.type = type;
     header.arg1 = arg1;
@@ -215,15 +215,7 @@ int udscs_write(struct udscs_connection *conn, uint32_t type, uint32_t arg1,
     memcpy(new_wbuf->buf, &header, sizeof(header));
     memcpy(new_wbuf->buf + sizeof(header), data, size);
 
-    if (conn->debug) {
-        if (type < conn->no_types)
-            syslog(LOG_DEBUG, "%p sent %s, arg1: %u, arg2: %u, size %u",
-                   conn, conn->type_to_string[type], arg1, arg2, size);
-        else
-            syslog(LOG_DEBUG,
-                   "%p sent invalid message %u, arg1: %u, arg2: %u, size %u",
-                   conn, type, arg1, arg2, size);
-    }
+    debug_print_message_header(conn, &header, "sent");
 
     if (conn->io_channel && conn->write_watch_id == 0)
         conn->write_watch_id =
@@ -234,7 +226,7 @@ int udscs_write(struct udscs_connection *conn, uint32_t type, uint32_t arg1,
 
     if (!conn->write_buf) {
         conn->write_buf = new_wbuf;
-        return 0;
+        return;
     }
 
     /* maybe we should limit the write_buf stack depth ? */
@@ -243,8 +235,6 @@ int udscs_write(struct udscs_connection *conn, uint32_t type, uint32_t arg1,
         wbuf = wbuf->next;
 
     wbuf->next = new_wbuf;
-
-    return 0;
 }
 
 /* A helper for udscs_do_read() */
@@ -252,18 +242,7 @@ static void udscs_read_complete(struct udscs_connection **connp)
 {
     struct udscs_connection *conn = *connp;
 
-    if (conn->debug) {
-        if (conn->header.type < conn->no_types)
-            syslog(LOG_DEBUG,
-                   "%p received %s, arg1: %u, arg2: %u, size %u",
-                   conn, conn->type_to_string[conn->header.type],
-                   conn->header.arg1, conn->header.arg2, conn->header.size);
-        else
-            syslog(LOG_DEBUG,
-               "%p received invalid message %u, arg1: %u, arg2: %u, size %u",
-               conn, conn->header.type, conn->header.arg1, conn->header.arg2,
-               conn->header.size);
-    }
+    debug_print_message_header(conn, &conn->header, "received");
 
     if (conn->read_callback) {
         conn->read_callback(connp, &conn->header, conn->data.buf);
@@ -271,7 +250,7 @@ static void udscs_read_complete(struct udscs_connection **connp)
             return;
     }
 
-    free(conn->data.buf);
+    g_free(conn->data.buf);
     memset(&conn->data, 0, sizeof(conn->data)); /* data.buf = NULL */
     conn->header_read = 0;
 }
@@ -312,12 +291,7 @@ static void udscs_do_read(struct udscs_connection **connp)
             }
             conn->data.pos = 0;
             conn->data.size = conn->header.size;
-            conn->data.buf = malloc(conn->data.size);
-            if (!conn->data.buf) {
-                syslog(LOG_ERR, "out of memory, disconnecting %p", conn);
-                udscs_destroy_connection(connp);
-                return;
-            }
+            conn->data.buf = g_malloc(conn->data.size);
         }
     } else {
         conn->data.pos += n;
@@ -354,8 +328,8 @@ static void udscs_do_write(struct udscs_connection **connp)
     wbuf->pos += n;
     if (wbuf->pos == wbuf->size) {
         conn->write_buf = wbuf->next;
-        free(wbuf->buf);
-        free(wbuf);
+        g_free(wbuf->buf);
+        g_free(wbuf);
     }
 }
 
@@ -367,12 +341,14 @@ static gboolean udscs_io_channel_cb(GIOChannel *source,
 
     if (condition & G_IO_IN) {
         udscs_do_read(&conn);
+        // coverity[check_after_deref] previous function can change conn
         if (conn == NULL)
             return G_SOURCE_REMOVE;
         return G_SOURCE_CONTINUE;
     }
     if (condition & G_IO_OUT) {
         udscs_do_write(&conn);
+        // coverity[check_after_deref] previous function can change conn
         if (conn == NULL)
             return G_SOURCE_REMOVE;
         if (conn->write_buf)
@@ -392,8 +368,6 @@ static gboolean udscs_io_channel_cb(GIOChannel *source,
 
 struct udscs_server {
     int fd;
-    const char * const *type_to_string;
-    int no_types;
     int debug;
     struct udscs_connection connections_head;
     udscs_connect_callback connect_callback;
@@ -405,7 +379,7 @@ struct udscs_server *udscs_create_server_for_fd(int fd,
     udscs_connect_callback connect_callback,
     udscs_read_callback read_callback,
     udscs_disconnect_callback disconnect_callback,
-    const char * const type_to_string[], int no_types, int debug)
+    int debug)
 {
     struct udscs_server *server;
 
@@ -414,12 +388,7 @@ struct udscs_server *udscs_create_server_for_fd(int fd,
         return NULL;
     }
 
-    server = calloc(1, sizeof(*server));
-    if (!server)
-        return NULL;
-
-    server->type_to_string = type_to_string;
-    server->no_types = no_types;
+    server = g_new0(struct udscs_server, 1);
     server->debug = debug;
     server->fd = fd;
     server->connect_callback = connect_callback;
@@ -433,7 +402,7 @@ struct udscs_server *udscs_create_server(const char *socketname,
     udscs_connect_callback connect_callback,
     udscs_read_callback read_callback,
     udscs_disconnect_callback disconnect_callback,
-    const char * const type_to_string[], int no_types, int debug)
+    int debug)
 {
     int c;
     int fd;
@@ -463,8 +432,7 @@ struct udscs_server *udscs_create_server(const char *socketname,
     }
 
     server = udscs_create_server_for_fd(fd, connect_callback, read_callback,
-                                        disconnect_callback, type_to_string,
-                                        no_types, debug);
+                                        disconnect_callback, debug);
 
     if (!server) {
         close(fd);
@@ -487,12 +455,12 @@ void udscs_destroy_server(struct udscs_server *server)
         conn = next_conn;
     }
     close(server->fd);
-    free(server);
+    g_free(server);
 }
 
-struct ucred udscs_get_peer_cred(struct udscs_connection *conn)
+int udscs_get_peer_pid(struct udscs_connection *conn)
 {
-    return conn->peer_cred;
+    return (int)conn->peer_cred.pid;
 }
 
 static void udscs_server_accept(struct udscs_server *server) {
@@ -509,16 +477,8 @@ static void udscs_server_accept(struct udscs_server *server) {
         return;
     }
 
-    new_conn = calloc(1, sizeof(*conn));
-    if (!new_conn) {
-        syslog(LOG_ERR, "out of memory, disconnecting new client");
-        close(fd);
-        return;
-    }
-
+    new_conn = g_new0(struct udscs_connection, 1);
     new_conn->fd = fd;
-    new_conn->type_to_string = server->type_to_string;
-    new_conn->no_types = server->no_types;
     new_conn->debug = server->debug;
     new_conn->read_callback = server->read_callback;
     new_conn->disconnect_callback = server->disconnect_callback;
@@ -528,7 +488,7 @@ static void udscs_server_accept(struct udscs_server *server) {
     if (r != 0) {
         syslog(LOG_ERR, "Could not get peercred, disconnecting new client");
         close(fd);
-        free(new_conn);
+        g_free(new_conn);
         return;
     }
 
@@ -541,7 +501,7 @@ static void udscs_server_accept(struct udscs_server *server) {
 
     if (server->debug)
         syslog(LOG_DEBUG, "new client accepted: %p, pid: %d",
-               new_conn, (int)new_conn->peer_cred.pid);
+               new_conn, udscs_get_peer_pid(new_conn));
 
     if (server->connect_callback)
         server->connect_callback(new_conn);
@@ -600,7 +560,7 @@ void udscs_server_handle_fds(struct udscs_server *server, fd_set *readfds,
     }
 }
 
-int udscs_server_write_all(struct udscs_server *server,
+void udscs_server_write_all(struct udscs_server *server,
         uint32_t type, uint32_t arg1, uint32_t arg2,
         const uint8_t *data, uint32_t size)
 {
@@ -608,12 +568,9 @@ int udscs_server_write_all(struct udscs_server *server,
 
     conn = server->connections_head.next;
     while (conn) {
-        if (udscs_write(conn, type, arg1, arg2, data, size))
-            return -1;
+        udscs_write(conn, type, arg1, arg2, data, size);
         conn = conn->next;
     }
-
-    return 0;
 }
 
 int udscs_server_for_all_clients(struct udscs_server *server,
