@@ -61,15 +61,12 @@ typedef void (*udscs_disconnect_callback)(struct udscs_connection *conn);
  * Only sockets bound to a pathname are supported.
  *
  * If debug is true then the events on this connection will be traced.
- * This includes the incoming and outgoing message names. So when debug is true
- * no_types must be set to the value of the highest valid message id + 1,
- * and type_to_string must point to a string array of size no_types for
- * converting the message ids to their names.
+ * This includes the incoming and outgoing message names.
  */
 struct udscs_connection *udscs_connect(const char *socketname,
     udscs_read_callback read_callback,
     udscs_disconnect_callback disconnect_callback,
-    const char * const type_to_string[], int no_types, int debug);
+    int debug);
 
 /* Close the connection, releases the corresponding resources and
  * sets *connp to NULL.
@@ -79,9 +76,8 @@ struct udscs_connection *udscs_connect(const char *socketname,
 void udscs_destroy_connection(struct udscs_connection **connp);
 
 /* Queue a message for delivery to the client connected through conn.
- * Return value: 0 on success -1 on error (only happens when malloc fails).
  */
-int udscs_write(struct udscs_connection *conn, uint32_t type, uint32_t arg1,
+void udscs_write(struct udscs_connection *conn, uint32_t type, uint32_t arg1,
         uint32_t arg2, const uint8_t *data, uint32_t size);
 
 /* Associates the specified user data with the connection. */
@@ -113,7 +109,7 @@ struct udscs_server *udscs_create_server_for_fd(int fd,
     udscs_connect_callback connect_callback,
     udscs_read_callback read_callback,
     udscs_disconnect_callback disconnect_callback,
-    const char * const type_to_string[], int no_types, int debug);
+    int debug);
 
 /* Create the unix domain socket specified by socketname and
  * start listening on it.
@@ -121,16 +117,13 @@ struct udscs_server *udscs_create_server_for_fd(int fd,
  *
  * If debug is true then the events on this socket and related individual
  * connections will be traced.
- * This includes the incoming and outgoing message names. So when debug is true
- * no_types must be set to the value of the highest valid message id + 1,
- * and type_to_string must point to a string array of size no_types for
- * converting the message ids to their names.
+ * This includes the incoming and outgoing message names.
  */
 struct udscs_server *udscs_create_server(const char *socketname,
     udscs_connect_callback connect_callback,
     udscs_read_callback read_callback,
     udscs_disconnect_callback disconnect_callback,
-    const char * const type_to_string[], int no_types, int debug);
+    int debug);
 
 /* Close all the server's connections and releases the corresponding
  * resources.
@@ -141,7 +134,7 @@ void udscs_destroy_server(struct udscs_server *server);
 /* Like udscs_write, but then send the message to all clients connected to
  * the server.
  */
-int udscs_server_write_all(struct udscs_server *server,
+void udscs_server_write_all(struct udscs_server *server,
     uint32_t type, uint32_t arg1, uint32_t arg2,
     const uint8_t *data, uint32_t size);
 
@@ -171,8 +164,8 @@ int udscs_server_fill_fds(struct udscs_server *server, fd_set *readfds,
 void udscs_server_handle_fds(struct udscs_server *server, fd_set *readfds,
     fd_set *writefds);
 
-/* Returns the peer's user credentials. */
-struct ucred udscs_get_peer_cred(struct udscs_connection *conn);
+/* Returns the peer's PID. */
+int udscs_get_peer_pid(struct udscs_connection *conn);
 
 #endif
 
