@@ -20,6 +20,7 @@
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
+#include <config.h>
 
 #include <glib.h>
 #include <glib/gstdio.h>
@@ -30,13 +31,14 @@
 
 #include <X11/extensions/Xinerama.h>
 
+#include "device-info.h"
 #include "vdagentd-proto.h"
 #include "x11.h"
 #include "x11-priv.h"
 
 #define MM_PER_INCH (25.4)
 
-static int error_handler(Display *display, XErrorEvent *error)
+static int ignore_error_handler(Display *display, XErrorEvent *error)
 {
     vdagent_x11_caught_error = 1;
     return 0;
@@ -81,18 +83,15 @@ static void free_randr_resources(struct vdagent_x11 *x11)
         for (i = 0 ; i < x11->randr.res->noutput; ++i) {
             XRRFreeOutputInfo(x11->randr.outputs[i]);
         }
-        free(x11->randr.outputs);
+        g_clear_pointer(&x11->randr.outputs, g_free);
     }
     if (x11->randr.crtcs != NULL) {
         for (i = 0 ; i < x11->randr.res->ncrtc; ++i) {
             XRRFreeCrtcInfo(x11->randr.crtcs[i]);
         }
-        free(x11->randr.crtcs);
+        g_clear_pointer(&x11->randr.crtcs, g_free);
     }
-    XRRFreeScreenResources(x11->randr.res);
-    x11->randr.res = NULL;
-    x11->randr.outputs = NULL;
-    x11->randr.crtcs = NULL;
+    g_clear_pointer(&x11->randr.res, XRRFreeScreenResources);
     x11->randr.num_monitors = 0;
 }
 
@@ -105,8 +104,8 @@ static void update_randr_res(struct vdagent_x11 *x11, int poll)
         x11->randr.res = XRRGetScreenResources(x11->display, x11->root_window[0]);
     else
         x11->randr.res = XRRGetScreenResourcesCurrent(x11->display, x11->root_window[0]);
-    x11->randr.outputs = malloc(x11->randr.res->noutput * sizeof(*x11->randr.outputs));
-    x11->randr.crtcs = malloc(x11->randr.res->ncrtc * sizeof(*x11->randr.crtcs));
+    x11->randr.outputs = g_new(XRROutputInfo *, x11->randr.res->noutput);
+    x11->randr.crtcs = g_new(XRRCrtcInfo *, x11->randr.res->ncrtc);
     for (i = 0 ; i < x11->randr.res->noutput; ++i) {
         x11->randr.outputs[i] = XRRGetOutputInfo(x11->display, x11->randr.res,
                                                  x11->randr.res->outputs[i]);
@@ -123,7 +122,7 @@ static void update_randr_res(struct vdagent_x11 *x11, int poll)
                               &x11->randr.min_height,
                               &x11->randr.max_width,
                               &x11->randr.max_height) != 1) {
-        syslog(LOG_ERR, "update_randr_res: RRGetScreenSizeRange failed");
+        syslog(LOG_ERR, "update_randr_res: XRRGetScreenSizeRange failed");
     }
 }
 
@@ -189,14 +188,14 @@ find_mode_by_name (struct vdagent_x11 *x11, char *name)
 }
 
 static XRRModeInfo *
-find_mode_by_size (struct vdagent_x11 *x11, int output, int width, int height)
+find_mode_by_size (struct vdagent_x11 *x11, int output_index, int width, int height)
 {
     int        	m;
     XRRModeInfo        *ret = NULL;
 
-    for (m = 0; m < x11->randr.outputs[output]->nmode; m++) {
+    for (m = 0; m < x11->randr.outputs[output_index]->nmode; m++) {
         XRRModeInfo *mode = mode_from_id(x11,
-                                         x11->randr.outputs[output]->modes[m]);
+                                         x11->randr.outputs[output_index]->modes[m]);
         if (mode && mode->width == width && mode->height == height) {
             ret = mode;
             break;
@@ -232,7 +231,7 @@ static void delete_mode(struct vdagent_x11 *x11, int output_index,
             break;
     }
     if (m < x11->randr.res->nmode) {
-        vdagent_x11_set_error_handler(x11, error_handler);
+        vdagent_x11_set_error_handler(x11, ignore_error_handler);
         XRRDeleteOutputMode (x11->display, x11->randr.res->outputs[output_index],
                              mode->id);
         XRRDestroyMode (x11->display, mode->id);
@@ -332,7 +331,7 @@ static XRRModeInfo *create_new_mode(struct vdagent_x11 *x11, int output_index,
     set_reduced_cvt_mode(&mode, width, height);
     mode.modeFlags = 0;
     mode.id = 0;
-    vdagent_x11_set_error_handler(x11, error_handler);
+    vdagent_x11_set_error_handler(x11, ignore_error_handler);
     XRRCreateMode (x11->display, x11->root_window[0], &mode);
     // ignore race error, if mode is created by others
     vdagent_x11_restore_error_handler(x11);
@@ -343,40 +342,48 @@ static XRRModeInfo *create_new_mode(struct vdagent_x11 *x11, int output_index,
     return find_mode_by_name(x11, modename);
 }
 
-static int xrandr_add_and_set(struct vdagent_x11 *x11, int output, int x, int y,
+static int xrandr_add_and_set(struct vdagent_x11 *x11, int output_index, int x, int y,
                               int width, int height)
 {
     XRRModeInfo *mode;
     int xid;
     Status s;
     RROutput outputs[1];
-    int old_width  = x11->randr.monitor_sizes[output].width;
-    int old_height = x11->randr.monitor_sizes[output].height;
+    int old_width;
+    int old_height;
 
-    if (!x11->randr.res || output >= x11->randr.res->noutput || output < 0) {
-        syslog(LOG_ERR, "%s: program error: missing RANDR or bad output",
-               __FUNCTION__);
+    if (!x11->randr.res) {
+        syslog(LOG_ERR, "%s: program error: missing RANDR", __FUNCTION__);
         return 0;
     }
+
+    if (output_index < 0 || output_index >= x11->randr.res->noutput) {
+        syslog(LOG_ERR, "%s: program error: bad output", __FUNCTION__);
+        return 0;
+    }
+
+    old_width  = x11->randr.monitor_sizes[output_index].width;
+    old_height = x11->randr.monitor_sizes[output_index].height;
+
     if (x11->set_crtc_config_not_functional) {
         /* fail, set_best_mode will find something close. */
         return 0;
     }
-    xid = x11->randr.res->outputs[output];
-    mode = find_mode_by_size(x11, output, width, height);
+    xid = x11->randr.res->outputs[output_index];
+    mode = find_mode_by_size(x11, output_index, width, height);
     if (!mode) {
-        mode = create_new_mode(x11, output, width, height);
+        mode = create_new_mode(x11, output_index, width, height);
     }
     if (!mode) {
         syslog(LOG_ERR, "failed to add a new mode");
         return 0;
     }
     XRRAddOutputMode(x11->display, xid, mode->id);
-    x11->randr.monitor_sizes[output].width = width;
-    x11->randr.monitor_sizes[output].height = height;
+    x11->randr.monitor_sizes[output_index].width = width;
+    x11->randr.monitor_sizes[output_index].height = height;
     outputs[0] = xid;
-    vdagent_x11_set_error_handler(x11, error_handler);
-    s = XRRSetCrtcConfig(x11->display, x11->randr.res, x11->randr.res->crtcs[output],
+    vdagent_x11_set_error_handler(x11, ignore_error_handler);
+    s = XRRSetCrtcConfig(x11->display, x11->randr.res, x11->randr.res->crtcs[output_index],
                          CurrentTime, x, y, mode->id, RR_Rotate_0, outputs,
                          1);
     if (vdagent_x11_restore_error_handler(x11) || (s != RRSetConfigSuccess)) {
@@ -387,33 +394,63 @@ static int xrandr_add_and_set(struct vdagent_x11 *x11, int output, int x, int y,
 
     /* clean the previous name, if any */
     if (width != old_width || height != old_height)
-        delete_mode(x11, output, old_width, old_height);
+        delete_mode(x11, output_index, old_width, old_height);
 
     return 1;
 }
 
-static void xrandr_disable_output(struct vdagent_x11 *x11, int output)
+// Looks up the xrandr output id associated with the given spice display id
+static RROutput get_xrandr_output_for_display_id(struct vdagent_x11 *x11, int display_id)
+{
+    guint map_size = g_hash_table_size(x11->guest_output_map);
+    if (map_size == 0) {
+        // we never got a device info message from the server, so just use old
+        // assumptions that the spice display id is equal to the index into the
+        // array of xrandr outputs
+        if (display_id < x11->randr.res->noutput) {
+            return x11->randr.res->outputs[display_id];
+        }
+    } else {
+        gpointer value;
+        if (g_hash_table_lookup_extended(x11->guest_output_map, GINT_TO_POINTER(display_id),
+                                         NULL, &value)) {
+            return *(gint64*)value;
+        }
+    }
+
+    // unable to find a valid output id
+    return -1;
+}
+
+static void xrandr_disable_nth_output(struct vdagent_x11 *x11, int output_index)
 {
     Status s;
 
-    if (!x11->randr.res || output >= x11->randr.res->noutput || output < 0) {
+    if (!x11->randr.res || output_index >= x11->randr.res->noutput || output_index < 0) {
         syslog(LOG_ERR, "%s: program error: missing RANDR or bad output",
                __FUNCTION__);
         return;
     }
 
+    XRROutputInfo *oinfo = x11->randr.outputs[output_index];
+    if (oinfo->ncrtc == 0) {
+        syslog(LOG_WARNING, "Output index %i doesn't have any associated CRTCs", output_index);
+        return;
+    }
+
+    // assume output only has a single crtc
     s = XRRSetCrtcConfig(x11->display, x11->randr.res,
-                         x11->randr.res->crtcs[output],
+                         oinfo->crtcs[0],
                          CurrentTime, 0, 0, None, RR_Rotate_0,
                          NULL, 0);
 
     if (s != RRSetConfigSuccess)
         syslog(LOG_ERR, "failed to disable monitor");
 
-    delete_mode(x11, output, x11->randr.monitor_sizes[output].width,
-                             x11->randr.monitor_sizes[output].height);
-    x11->randr.monitor_sizes[output].width  = 0;
-    x11->randr.monitor_sizes[output].height = 0;
+    delete_mode(x11, output_index, x11->randr.monitor_sizes[output_index].width,
+                             x11->randr.monitor_sizes[output_index].height);
+    x11->randr.monitor_sizes[output_index].width  = 0;
+    x11->randr.monitor_sizes[output_index].height = 0;
 }
 
 static int set_screen_to_best_size(struct vdagent_x11 *x11, int width, int height,
@@ -652,42 +689,69 @@ static int config_size(int num_of_monitors)
                            num_of_monitors * sizeof(VDAgentMonConfig);
 }
 
+// gets monitor information about the specified output index and returns true if there was no error
+static bool get_monitor_info_for_output_index(struct vdagent_x11 *x11, int output_index,
+                                              int *x, int *y, int *width, int *height)
+{
+    g_return_val_if_fail (output_index < x11->randr.res->noutput, false);
+    g_return_val_if_fail (x != NULL, false);
+    g_return_val_if_fail (y != NULL, false);
+    g_return_val_if_fail (width != NULL, false);
+    g_return_val_if_fail (height != NULL, false);
+
+    int j;
+    XRRCrtcInfo *crtc = NULL;
+    XRRModeInfo *mode;
+
+    if (x11->randr.outputs[output_index]->ncrtc == 0)
+        goto zeroed; /* Monitor disabled */
+
+    for (j = 0; crtc == NULL && j < x11->randr.outputs[output_index]->ncrtc; j++) {
+        crtc = crtc_from_id(x11, x11->randr.outputs[output_index]->crtcs[j]);
+    }
+    if (!crtc) {
+        // error. stale xrandr info?
+        return false;
+    }
+
+    mode = mode_from_id(x11, crtc->mode);
+    if (!mode)
+        goto zeroed; /* monitor disabled */
+
+    *x = crtc->x;
+    *y = crtc->y;
+    *width = mode->width;
+    *height = mode->height;
+    return true;
+
+zeroed:
+    *x = 0;
+    *y = 0;
+    *width = 0;
+    *height = 0;
+    return true;
+}
+
 static VDAgentMonitorsConfig *get_current_mon_config(struct vdagent_x11 *x11)
 {
     int i, num_of_monitors = 0;
-    XRRModeInfo *mode;
     XRRScreenResources *res = x11->randr.res;
     VDAgentMonitorsConfig *mon_config;
 
-    mon_config = calloc(1, config_size(res->noutput));
-    if (!mon_config) {
-        syslog(LOG_ERR, "out of memory allocating current monitor config");
-        return NULL;
-    }
+    mon_config = g_malloc0(config_size(res->noutput));
 
     for (i = 0 ; i < res->noutput; i++) {
-        int j;
-        XRRCrtcInfo *crtc = NULL;
-
-        if (x11->randr.outputs[i]->ncrtc == 0)
-            continue; /* Monitor disabled, already zero-ed by calloc */
-        if (x11->randr.outputs[i]->crtc == 0)
-            continue; /* Monitor disabled */
-
-        for (j = 0; crtc == NULL && j < x11->randr.outputs[i]->ncrtc; j++) {
-            crtc = crtc_from_id(x11, x11->randr.outputs[i]->crtcs[j]);
-        }
-        if (!crtc)
+        int x, y, width, height;
+        if (!get_monitor_info_for_output_index(x11, i, &x, &y, &width, &height)) {
+            syslog(LOG_WARNING, "Unable to get monitor info for output id %d", i);
             goto error;
+        }
 
-        mode = mode_from_id(x11, crtc->mode);
-        if (!mode)
-            continue; /* Monitor disabled, already zero-ed by calloc */
-
-        mon_config->monitors[i].x      = crtc->x;
-        mon_config->monitors[i].y      = crtc->y;
-        mon_config->monitors[i].width  = mode->width;
-        mon_config->monitors[i].height = mode->height;
+        VDAgentMonConfig *mon = &mon_config->monitors[i];
+        mon->x = x;
+        mon->y = y;
+        mon->width = width;
+        mon->height = height;
         num_of_monitors = i + 1;
     }
     mon_config->num_of_monitors = num_of_monitors;
@@ -696,7 +760,7 @@ static VDAgentMonitorsConfig *get_current_mon_config(struct vdagent_x11 *x11)
 
 error:
     syslog(LOG_ERR, "error: inconsistent or stale data from X");
-    free(mon_config);
+    g_free(mon_config);
     return NULL;
 }
 
@@ -707,17 +771,90 @@ static void dump_monitors_config(struct vdagent_x11 *x11,
     int i;
     VDAgentMonConfig *m;
 
-    syslog(LOG_DEBUG, "%s: %d, %x", prefix, mon_config->num_of_monitors,
+    syslog(LOG_DEBUG, "Monitors config %s: %d, %x", prefix, mon_config->num_of_monitors,
            mon_config->flags);
     for (i = 0 ; i < mon_config->num_of_monitors; ++i) {
         m = &mon_config->monitors[i];
         if (!monitor_enabled(m))
             continue;
-        syslog(LOG_DEBUG, "received monitor %d config %dx%d+%d+%d", i,
-               m->width, m->height, m->x, m->y);
+        syslog(LOG_DEBUG, "    monitor %d, config %dx%d+%d+%d",
+               i, m->width, m->height, m->x, m->y);
     }
 }
 
+// handle the device info message from the server. This will allow us to
+// maintain a mapping from spice display id to xrandr output
+void vdagent_x11_handle_graphics_device_info(struct vdagent_x11 *x11, uint8_t *data, size_t size)
+{
+    VDAgentGraphicsDeviceInfo *graphics_device_info = (VDAgentGraphicsDeviceInfo *)data;
+    VDAgentDeviceDisplayInfo *device_display_info = graphics_device_info->display_info;
+
+    void *buffer_end = data + size;
+
+    syslog(LOG_INFO, "Received Graphics Device Info:");
+
+    for (size_t i = 0; i < graphics_device_info->count; ++i) {
+        if ((void*) device_display_info > buffer_end ||
+                (void*) (&device_display_info->device_address +
+                    device_display_info->device_address_len) > buffer_end) {
+            syslog(LOG_ERR, "Malformed graphics_display_info message, "
+                   "extends beyond the end of the buffer");
+            break;
+        }
+
+        // make sure the string is terminated:
+        if (device_display_info->device_address_len > 0) {
+            device_display_info->device_address[device_display_info->device_address_len - 1] = '\0';
+        } else {
+            syslog(LOG_WARNING, "Zero length device_address received for channel_id: %u, monitor_id: %u",
+                   device_display_info->channel_id, device_display_info->monitor_id);
+        }
+
+        RROutput x_output;
+        if (lookup_xrandr_output_for_device_info(device_display_info, x11->display,
+                                                 x11->randr.res, &x_output)) {
+            gint64 *value = g_new(gint64, 1);
+            *value = x_output;
+
+            syslog(LOG_INFO, "Adding graphics device info: channel_id: %u monitor_id: "
+                   "%u device_address: %s, device_display_id: %u xrandr output ID: %lu",
+                   device_display_info->channel_id,
+                   device_display_info->monitor_id,
+                   device_display_info->device_address,
+                   device_display_info->device_display_id,
+                   x_output);
+
+            g_hash_table_insert(x11->guest_output_map,
+                GUINT_TO_POINTER(device_display_info->channel_id + device_display_info->monitor_id),
+                value);
+        } else {
+            syslog(LOG_INFO, "channel_id: %u monitor_id: %u device_address: %s, "
+                   "device_display_id: %u xrandr output ID NOT FOUND",
+                   device_display_info->channel_id,
+                   device_display_info->monitor_id,
+                   device_display_info->device_address,
+                   device_display_info->device_display_id);
+        }
+
+        device_display_info = (VDAgentDeviceDisplayInfo*) ((char*) device_display_info +
+            sizeof(VDAgentDeviceDisplayInfo) + device_display_info->device_address_len);
+    }
+
+    // make sure daemon is up-to-date with (possibly updated) device IDs
+    vdagent_x11_send_daemon_guest_xorg_res(x11, 1);
+}
+
+static int get_output_index_for_display_id(struct vdagent_x11 *x11, int display_id)
+{
+    RROutput oid = get_xrandr_output_for_display_id(x11, display_id);
+    for (int i = 0; i < x11->randr.res->noutput; i++) {
+        if (oid == x11->randr.res->outputs[i]) {
+            return i;
+        }
+    }
+    return -1;
+}
+
 /*
  * Set monitor configuration according to client request.
  *
@@ -793,20 +930,40 @@ void vdagent_x11_set_monitor_config(struct vdagent_x11 *x11,
     g_unlink(config);
     g_free(config);
 
-    for (i = mon_config->num_of_monitors; i < x11->randr.res->noutput; i++)
-        xrandr_disable_output(x11, i);
+    // disable all outputs that don't have associated entries in the MonitorConfig
+    for (i = 0; i < x11->randr.res->noutput; i++) {
+        bool disable = true;
+        // check if this xrandr output is represented by an item in mon_config
+        for (int j = 0; j < mon_config->num_of_monitors; j++) {
+            // j represents the display id of an enabled monitor. Check whether
+            // an enabled xrandr output is represented by this id.
+            RROutput oid = get_xrandr_output_for_display_id(x11, j);
+            if (oid == x11->randr.res->outputs[i]) {
+                disable = false;
+            }
+        }
+        if (disable) {
+            xrandr_disable_nth_output(x11, i);
+        }
+    }
 
-    /* First, disable disabled CRTCs... */
+    /* disable CRTCs that are present but explicitly disabled in the
+     * MonitorConfig */
     for (i = 0; i < mon_config->num_of_monitors; ++i) {
         if (!monitor_enabled(&mon_config->monitors[i])) {
-            xrandr_disable_output(x11, i);
+            int output_index = get_output_index_for_display_id(x11, i);
+            if (output_index != -1) {
+                xrandr_disable_nth_output(x11, output_index);
+            } else {
+                syslog(LOG_WARNING, "Unable to find a guest output index for spice display %i", i);
+            }
         }
     }
 
     /* ... and disable the ones that would be bigger than
      * the new RandR screen once it is resized. If they are enabled the
      * XRRSetScreenSize call will fail with BadMatch. They will be
-     * re-enabled after hanging the screen size.
+     * re-enabled after changing the screen size.
      */
     for (i = 0; i < curr->num_of_monitors; ++i) {
         int width, height;
@@ -819,11 +976,15 @@ void vdagent_x11_set_monitor_config(struct vdagent_x11 *x11,
 
         if ((x + width > primary_w) || (y + height > primary_h)) {
             if (x11->debug)
-                syslog(LOG_DEBUG, "Disabling monitor %d: "
-                       "%dx%d+%d+%d > (%d,%d)",
+                syslog(LOG_DEBUG, "Disabling monitor %d: %dx%d+%d+%d > (%d,%d)",
                        i, width, height, x, y, primary_w, primary_h);
 
-            xrandr_disable_output(x11, i);
+            int output_index = get_output_index_for_display_id(x11, i);
+            if (output_index != -1) {
+                xrandr_disable_nth_output(x11, output_index);
+            } else {
+                syslog(LOG_WARNING, "Unable to find a guest output index for spice display %i", i);
+            }
         }
     }
 
@@ -836,7 +997,7 @@ void vdagent_x11_set_monitor_config(struct vdagent_x11 *x11,
         if (x11->debug)
             syslog(LOG_DEBUG, "Changing screen size to %dx%d",
                    primary_w, primary_h);
-        vdagent_x11_set_error_handler(x11, error_handler);
+        vdagent_x11_set_error_handler(x11, ignore_error_handler);
         XRRSetScreenSize(x11->display, x11->root_window[0], primary_w, primary_h,
                          width_mm, height_mm);
         if (vdagent_x11_restore_error_handler(x11)) {
@@ -844,15 +1005,12 @@ void vdagent_x11_set_monitor_config(struct vdagent_x11 *x11,
             if (!fallback) {
                 syslog(LOG_WARNING, "Restoring previous config");
                 vdagent_x11_set_monitor_config(x11, curr, 1);
-                free(curr);
+                g_free(curr);
                 /* Remember this config failed, if the client is maximized or
                    fullscreen it will keep sending the failing config. */
-                free(x11->randr.failed_conf);
+                g_free(x11->randr.failed_conf);
                 x11->randr.failed_conf =
-                    malloc(config_size(mon_config->num_of_monitors));
-                if (x11->randr.failed_conf)
-                    memcpy(x11->randr.failed_conf, mon_config,
-                           config_size(mon_config->num_of_monitors));
+                    g_memdup(mon_config, config_size(mon_config->num_of_monitors));
                 return;
             }
         }
@@ -872,11 +1030,22 @@ void vdagent_x11_set_monitor_config(struct vdagent_x11 *x11,
         height = mon_config->monitors[i].height;
         x = mon_config->monitors[i].x;
         y = mon_config->monitors[i].y;
-        if (!xrandr_add_and_set(x11, i, x, y, width, height) &&
+
+        if (x11->debug) {
+            syslog(LOG_DEBUG, "Setting resolution for monitor %d: %dx%d+%d+%d)",
+                   i, width, height, x, y);
+        }
+
+        int output_index = get_output_index_for_display_id(x11, i);
+        if (output_index != -1) {
+            if (!xrandr_add_and_set(x11, output_index, x, y, width, height) &&
                 enabled_monitors(mon_config) == 1) {
-            set_screen_to_best_size(x11, width, height,
-                                    &primary_w, &primary_h);
-            break;
+                set_screen_to_best_size(x11, width, height,
+                                        &primary_w, &primary_h);
+                break;
+            }
+        } else {
+            syslog(LOG_WARNING, "Unable to find a guest output index for spice display %i", i);
         }
     }
 
@@ -895,38 +1064,52 @@ exit:
 
     /* Flush output buffers and consume any pending events */
     vdagent_x11_do_read(x11);
-    free(curr);
+    g_free(curr);
 }
 
 void vdagent_x11_send_daemon_guest_xorg_res(struct vdagent_x11 *x11, int update)
 {
-    struct vdagentd_guest_xorg_resolution *res = NULL;
+    GArray *res_array = g_array_new(FALSE, FALSE, sizeof(struct vdagentd_guest_xorg_resolution));
     int i, width = 0, height = 0, screen_count = 0;
 
     if (x11->has_xrandr) {
-        VDAgentMonitorsConfig *curr;
-
         if (update)
             update_randr_res(x11, 0);
 
-        curr = get_current_mon_config(x11);
-        if (!curr)
-            goto no_info;
-
-        screen_count = curr->num_of_monitors;
-        res = malloc(screen_count * sizeof(*res));
-        if (!res) {
-            free(curr);
-            goto no_mem;
-        }
+        screen_count = x11->randr.res->noutput;
 
         for (i = 0; i < screen_count; i++) {
-            res[i].width  = curr->monitors[i].width;
-            res[i].height = curr->monitors[i].height;
-            res[i].x = curr->monitors[i].x;
-            res[i].y = curr->monitors[i].y;
+            struct vdagentd_guest_xorg_resolution curr;
+            if (!get_monitor_info_for_output_index(x11, i, &curr.x, &curr.y,
+                        &curr.width, &curr.height)) {
+                g_array_free(res_array, TRUE);
+                goto no_info;
+            }
+            if (g_hash_table_size(x11->guest_output_map) == 0) {
+                syslog(LOG_DEBUG, "No guest output map, using output index as display id");
+                curr.display_id = i;
+                g_array_append_val(res_array, curr);
+            } else {
+                // There may be multiple spice outputs representing a single guest output. Send them
+                // all down.
+                RROutput output_id = x11->randr.res->outputs[i];
+                GHashTableIter iter;
+                gpointer key, value;
+                g_hash_table_iter_init(&iter, x11->guest_output_map);
+                bool found = false;
+                while (g_hash_table_iter_next(&iter, &key, &value)) {
+                    gint64 *other_id = value;
+                    if (*other_id == output_id) {
+                        curr.display_id = GPOINTER_TO_INT(key);
+                        g_array_append_val(res_array, curr);
+                        found = true;
+                    }
+                }
+                if (!found) {
+                    syslog(LOG_WARNING, "Unable to find a display id for output index %d)", i);
+                }
+            }
         }
-        free(curr);
         width  = x11->width[0];
         height = x11->height[0];
     } else if (x11->has_xinerama) {
@@ -935,55 +1118,61 @@ void vdagent_x11_send_daemon_guest_xorg_res(struct vdagent_x11 *x11, int update)
         screen_info = XineramaQueryScreens(x11->display, &screen_count);
         if (!screen_info)
             goto no_info;
-        res = malloc(screen_count * sizeof(*res));
-        if (!res) {
-            XFree(screen_info);
-            goto no_mem;
-        }
+        g_array_set_size(res_array, screen_count);
         for (i = 0; i < screen_count; i++) {
             if (screen_info[i].screen_number >= screen_count) {
                 syslog(LOG_ERR, "Invalid screen number in xinerama screen info (%d >= %d)",
                        screen_info[i].screen_number, screen_count);
                 XFree(screen_info);
-                free(res);
+                g_array_free(res_array, true);
                 return;
             }
-            res[screen_info[i].screen_number].width = screen_info[i].width;
-            res[screen_info[i].screen_number].height = screen_info[i].height;
-            res[screen_info[i].screen_number].x = screen_info[i].x_org;
-            res[screen_info[i].screen_number].y = screen_info[i].y_org;
+            struct vdagentd_guest_xorg_resolution *curr = &g_array_index(res_array,
+                                                                         struct vdagentd_guest_xorg_resolution,
+                                                                         screen_info[i].screen_number);
+            curr->width = screen_info[i].width;
+            curr->height = screen_info[i].height;
+            curr->x = screen_info[i].x_org;
+            curr->y = screen_info[i].y_org;
         }
         XFree(screen_info);
         width  = x11->width[0];
         height = x11->height[0];
     } else {
 no_info:
-        screen_count = x11->screen_count;
-        res = malloc(screen_count * sizeof(*res));
-        if (!res)
-            goto no_mem;
         for (i = 0; i < screen_count; i++) {
-            res[i].width  = x11->width[i];
-            res[i].height = x11->height[i];
+            struct vdagentd_guest_xorg_resolution res;
+            res.width  = x11->width[i];
+            res.height = x11->height[i];
             /* No way to get screen coordinates, assume rtl order */
-            res[i].x = width;
-            res[i].y = 0;
+            res.x = width;
+            res.y = 0;
             width += x11->width[i];
             if (x11->height[i] > height)
                 height = x11->height[i];
+            g_array_append_val(res_array, res);
         }
     }
 
+    if (screen_count == 0) {
+        syslog(LOG_DEBUG, "Screen count is zero, are we on wayland?");
+        g_array_free(res_array, TRUE);
+        return;
+    }
+
     if (x11->debug) {
-        for (i = 0; i < screen_count; i++)
-            syslog(LOG_DEBUG, "Screen %d %dx%d%+d%+d", i, res[i].width,
-                   res[i].height, res[i].x, res[i].y);
+        syslog(LOG_DEBUG, "Sending guest screen resolutions to vdagentd:");
+        if (res_array->len > screen_count) {
+            syslog(LOG_DEBUG, "(NOTE: list may contain overlapping areas when multiple spice displays show the same guest output)");
+        }
+        for (i = 0; i < res_array->len; i++) {
+            struct vdagentd_guest_xorg_resolution *res = (struct vdagentd_guest_xorg_resolution*)res_array->data;
+            syslog(LOG_DEBUG, "   screen %d %dx%d%+d%+d, display_id=%d", i,
+                   res[i].width, res[i].height, res[i].x, res[i].y, res[i].display_id);
+        }
     }
 
     udscs_write(x11->vdagentd, VDAGENTD_GUEST_XORG_RESOLUTION, width, height,
-                (uint8_t *)res, screen_count * sizeof(*res));
-    free(res);
-    return;
-no_mem:
-    syslog(LOG_ERR, "out of memory while trying to send resolutions, not sending resolutions.");
+                (uint8_t *)res_array->data, res_array->len * sizeof(struct vdagentd_guest_xorg_resolution));
+    g_array_free(res_array, TRUE);
 }
