@@ -31,9 +31,7 @@
    Calling XPending when-ever we return to the mainloop also ensures any
    pending writes are flushed. */
 
-#ifdef HAVE_CONFIG_H
-# include <config.h>
-#endif
+#include <config.h>
 
 #include <glib.h>
 #ifdef WITH_GTK
@@ -210,19 +208,20 @@ struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
 #endif
     gchar *net_wm_name = NULL;
 
-    x11 = calloc(1, sizeof(*x11));
-    if (!x11) {
-        syslog(LOG_ERR, "out of memory allocating vdagent_x11 struct");
-        return NULL;
-    }
-
+    x11 = g_new0(struct vdagent_x11, 1);
     x11->vdagentd = vdagentd;
     x11->debug = debug;
 
+    x11->guest_output_map = g_hash_table_new_full(&g_direct_hash,
+                                                  &g_direct_equal,
+                                                  NULL,
+                                                  &g_free);
+
+
     x11->display = XOpenDisplay(NULL);
     if (!x11->display) {
         syslog(LOG_ERR, "could not connect to X-server");
-        free(x11);
+        g_free(x11);
         return NULL;
     }
 
@@ -231,7 +230,7 @@ struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
         syslog(LOG_ERR, "Error too much screens: %d > %d",
                x11->screen_count, MAX_SCREENS);
         XCloseDisplay(x11->display);
-        free(x11);
+        g_free(x11);
         return NULL;
     }
 
@@ -344,9 +343,10 @@ void vdagent_x11_destroy(struct vdagent_x11 *x11, int vdagentd_disconnected)
     }
 #endif
 
+    g_hash_table_destroy(x11->guest_output_map);
     XCloseDisplay(x11->display);
-    free(x11->randr.failed_conf);
-    free(x11);
+    g_free(x11->randr.failed_conf);
+    g_free(x11);
 }
 
 int vdagent_x11_get_fd(struct vdagent_x11 *x11)
@@ -400,6 +400,7 @@ static void vdagent_x11_set_clipboard_owner(struct vdagent_x11 *x11,
                 x11->selection_req_data_size = 0;
                 x11->selection_req_atom = None;
             } else {
+                // coverity[var_deref_op] if it is not the first there's a previous
                 prev_sel->next = next_sel;
             }
             free(curr_sel);
@@ -428,6 +429,7 @@ static void vdagent_x11_set_clipboard_owner(struct vdagent_x11 *x11,
                 x11->clipboard_data_size = 0;
                 x11->expect_property_notify = 0;
             } else {
+                // coverity[var_deref_op] if it is not the first there's a previous
                 prev_conv->next = next_conv;
             }
             free(curr_conv);
@@ -730,6 +732,7 @@ static int vdagent_x11_get_selection(struct vdagent_x11 *x11, XEvent *event,
         len *= sizeof(short);
         break;
     case 32:
+        // on X11, a format of 32 is always a long
         len *= sizeof(long);
         break;
     }
@@ -942,6 +945,12 @@ static void vdagent_x11_handle_targets_notify(struct vdagent_x11 *x11,
 
     x11->expected_targets_notifies[selection]--;
 
+    if (x11->ignore_targets_notifies[selection] > 0) {
+        x11->ignore_targets_notifies[selection]--;
+        VSELPRINTF("ignoring selection notify TARGETS");
+        return;
+    }
+
     /* If we have more targets_notifies pending, ignore this one, we
        are only interested in the targets list of the current owner
        (which is the last one we've requested a targets list from) */
@@ -1243,6 +1252,11 @@ void vdagent_x11_clipboard_grab(struct vdagent_x11 *x11, uint8_t selection,
                        x11->selection_window, CurrentTime);
     vdagent_x11_set_clipboard_owner(x11, selection, owner_client);
 
+    /* If there're pending requests for targets, ignore the returned
+     * targets as the XSetSelectionOwner() call above made them invalid */
+    x11->ignore_targets_notifies[selection] =
+        x11->expected_targets_notifies[selection];
+
     /* Flush output buffers and consume any pending events */
     vdagent_x11_do_read(x11);
 }
