@@ -18,6 +18,7 @@
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
+#include <config.h>
 
 #include "session-info.h"
 
@@ -45,11 +46,6 @@ char *session_info_session_for_pid(struct session_info *si, uint32_t pid)
     return NULL;
 }
 
-gboolean session_is_locked(struct session_info *ck)
-{
-    return FALSE;
-}
-
 gboolean session_info_is_user(struct session_info *si)
 {
     return TRUE;
