@@ -19,9 +19,7 @@
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
-#ifdef HAVE_CONFIG_H
 #include <config.h>
-#endif
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -42,12 +40,13 @@
 
 #include "udscs.h"
 #include "vdagentd-proto.h"
-#include "vdagentd-proto-strings.h"
 #include "uinput.h"
 #include "xorg-conf.h"
 #include "virtio-port.h"
 #include "session-info.h"
 
+#define DEFAULT_UINPUT_DEVICE "/dev/uinput"
+
 struct agent_data {
     char *session;
     int width;
@@ -57,13 +56,17 @@ struct agent_data {
 };
 
 /* variables */
-static const char *pidfilename = "/var/run/spice-vdagentd/spice-vdagentd.pid";
-static const char *portdev = DEFAULT_VIRTIO_PORT_PATH;
-static const char *vdagentd_socket = VDAGENTD_SOCKET;
-static const char *uinput_device = "/dev/uinput";
+static const char *pidfilename = "/run/spice-vdagentd/spice-vdagentd.pid";
+
+static gchar *portdev = NULL;
+static gchar *vdagentd_socket = NULL;
+static gchar *uinput_device = NULL;
 static int debug = 0;
-static int uinput_fake = 0;
-static int only_once = 0;
+static gboolean uinput_fake = FALSE;
+static gboolean only_once = FALSE;
+static gboolean do_daemonize = TRUE;
+static gboolean want_session_info = TRUE;
+
 static struct udscs_server *server = NULL;
 static struct vdagent_virtio_port *virtio_port = NULL;
 static GHashTable *active_xfers = NULL;
@@ -118,12 +121,7 @@ static void send_capabilities(struct vdagent_virtio_port *vport,
     uint32_t size;
 
     size = sizeof(*caps) + VD_AGENT_CAPS_BYTES;
-    caps = calloc(1, size);
-    if (!caps) {
-        syslog(LOG_ERR, "out of memory allocating capabilities array (write)");
-        return;
-    }
-
+    caps = g_malloc0(size);
     caps->request = request;
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_MOUSE_STATE);
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_MONITORS_CONFIG);
@@ -134,12 +132,13 @@ static void send_capabilities(struct vdagent_virtio_port *vport,
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_GUEST_LINEEND_LF);
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_MAX_CLIPBOARD);
     VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_AUDIO_VOLUME_SYNC);
+    VD_AGENT_SET_CAPABILITY(caps->caps, VD_AGENT_CAP_GRAPHICS_DEVICE_INFO);
     virtio_msg_uint32_to_le((uint8_t *)caps, size, 0);
 
     vdagent_virtio_port_write(vport, VDP_CLIENT_PORT,
                               VD_AGENT_ANNOUNCE_CAPABILITIES, 0,
                               (uint8_t *)caps, size);
-    free(caps);
+    g_free(caps);
 }
 
 static void do_client_disconnect(void)
@@ -190,16 +189,8 @@ static void do_client_monitors(struct vdagent_virtio_port *vport, int port_nr,
 
     vdagentd_write_xorg_conf(new_monitors);
 
-    if (!mon_config ||
-            mon_config->num_of_monitors != new_monitors->num_of_monitors) {
-        free(mon_config);
-        mon_config = malloc(size);
-        if (!mon_config) {
-            syslog(LOG_ERR, "out of memory allocating monitors config");
-            return;
-        }
-    }
-    memcpy(mon_config, new_monitors, size);
+    g_free(mon_config);
+    mon_config = g_memdup(new_monitors, size);
 
     /* Send monitor config to currently active agent */
     if (active_session_conn)
@@ -230,19 +221,10 @@ static void do_client_capabilities(struct vdagent_virtio_port *vport,
     VDAgentMessage *message_header,
     VDAgentAnnounceCapabilities *caps)
 {
-    int new_size = VD_AGENT_CAPS_SIZE_FROM_MSG_SIZE(message_header->size);
+    capabilities_size = VD_AGENT_CAPS_SIZE_FROM_MSG_SIZE(message_header->size);
+    g_free(capabilities);
+    capabilities = g_memdup(caps->caps, capabilities_size * sizeof(uint32_t));
 
-    if (capabilities_size != new_size) {
-        capabilities_size = new_size;
-        free(capabilities);
-        capabilities = malloc(capabilities_size * sizeof(uint32_t));
-        if (!capabilities) {
-            syslog(LOG_ERR, "oom allocating capabilities array (read)");
-            capabilities_size = 0;
-            return;
-        }
-    }
-    memcpy(capabilities, caps->caps, capabilities_size * sizeof(uint32_t));
     if (caps->request) {
         /* Report the previous client has disconnected. */
         do_client_disconnect();
@@ -322,7 +304,7 @@ static void send_file_xfer_status(struct vdagent_virtio_port *vport,
         data_size = 0;
     }
 
-    status = malloc(sizeof(*status) + data_size);
+    status = g_malloc(sizeof(*status) + data_size);
     status->id = GUINT32_TO_LE(id);
     status->result = GUINT32_TO_LE(xfer_status);
     if (data)
@@ -336,7 +318,7 @@ static void send_file_xfer_status(struct vdagent_virtio_port *vport,
                                   VD_AGENT_FILE_XFER_STATUS, 0,
                                   (uint8_t *)status, sizeof(*status) + data_size);
 
-    free(status);
+    g_free(status);
 }
 
 static void do_client_file_xfer(struct vdagent_virtio_port *vport,
@@ -392,6 +374,16 @@ static void do_client_file_xfer(struct vdagent_virtio_port *vport,
     udscs_write(conn, msg_type, 0, 0, data, message_header->size);
 }
 
+static void forward_data_to_session_agent(uint32_t type, uint8_t *data, size_t size)
+{
+    if (active_session_conn == NULL) {
+        syslog(LOG_DEBUG, "No active session, can't forward message (type %u)", type);
+        return;
+    }
+
+    udscs_write(active_session_conn, type, 0, 0, data, size);
+}
+
 static gsize vdagent_message_min_size[] =
 {
     -1, /* Does not exist */
@@ -410,6 +402,7 @@ static gsize vdagent_message_min_size[] =
     0, /* VD_AGENT_CLIENT_DISCONNECTED */
     sizeof(VDAgentMaxClipboard), /* VD_AGENT_MAX_CLIPBOARD */
     sizeof(VDAgentAudioVolumeSync), /* VD_AGENT_AUDIO_VOLUME_SYNC */
+    sizeof(VDAgentGraphicsDeviceInfo), /* VD_AGENT_GRAPHICS_DEVICE_INFO */
 };
 
 static void vdagent_message_clipboard_from_le(VDAgentMessage *message_header,
@@ -494,6 +487,7 @@ static gboolean vdagent_message_check_size(const VDAgentMessage *message_header)
     case VD_AGENT_CLIPBOARD_GRAB:
     case VD_AGENT_AUDIO_VOLUME_SYNC:
     case VD_AGENT_ANNOUNCE_CAPABILITIES:
+    case VD_AGENT_GRAPHICS_DEVICE_INFO:
         if (message_header->size < min_size) {
             syslog(LOG_ERR, "read: invalid message size: %u for message type: %u",
                    message_header->size, message_header->type);
@@ -521,6 +515,8 @@ static gboolean vdagent_message_check_size(const VDAgentMessage *message_header)
     return TRUE;
 }
 
+static VDAgentGraphicsDeviceInfo *device_info = NULL;
+static size_t device_info_size = 0;
 static int virtio_port_read_complete(
         struct vdagent_virtio_port *vport,
         int port_nr,
@@ -567,6 +563,18 @@ static int virtio_port_read_complete(
         syslog(LOG_DEBUG, "Set max clipboard: %d", max_clipboard);
         break;
     }
+    case VD_AGENT_GRAPHICS_DEVICE_INFO: {
+        if (device_info) {
+            g_free(device_info);
+            device_info = NULL;
+            device_info_size = 0;
+        }
+        // store device info for re-sending when a session agent reconnects
+        device_info = g_memdup(data, message_header->size);
+        device_info_size = message_header->size;
+        forward_data_to_session_agent(VDAGENTD_GRAPHICS_DEVICE_INFO, data, message_header->size);
+        break;
+    }
     case VD_AGENT_AUDIO_VOLUME_SYNC: {
         VDAgentAudioVolumeSync *vdata = (VDAgentAudioVolumeSync *)data;
         virtio_msg_uint16_from_le((uint8_t *)vdata, message_header->size,
@@ -843,16 +851,10 @@ static gboolean remove_active_xfers(gpointer key, gpointer value, gpointer conn)
 static void agent_connect(struct udscs_connection *conn)
 {
     struct agent_data *agent_data;
-
-    agent_data = calloc(1, sizeof(*agent_data));
-    if (!agent_data) {
-        syslog(LOG_ERR, "Out of memory allocating agent data, disconnecting");
-        udscs_destroy_connection(&conn);
-        return;
-    }
+    agent_data = g_new0(struct agent_data, 1);
 
     if (session_info) {
-        uint32_t pid = udscs_get_peer_cred(conn).pid;
+        uint32_t pid = udscs_get_peer_pid(conn);
         agent_data->session = session_info_session_for_pid(session_info, pid);
     }
 
@@ -860,6 +862,11 @@ static void agent_connect(struct udscs_connection *conn)
     udscs_write(conn, VDAGENTD_VERSION, 0, 0,
                 (uint8_t *)VERSION, strlen(VERSION) + 1);
     update_active_session_connection(conn);
+
+    if (device_info) {
+        forward_data_to_session_agent(VDAGENTD_GRAPHICS_DEVICE_INFO,
+                                      (uint8_t *) device_info, device_info_size);
+    }
 }
 
 static void agent_disconnect(struct udscs_connection *conn)
@@ -868,55 +875,81 @@ static void agent_disconnect(struct udscs_connection *conn)
 
     g_hash_table_foreach_remove(active_xfers, remove_active_xfers, conn);
 
-    free(agent_data->session);
-    agent_data->session = NULL;
+    g_clear_pointer(&agent_data->session, g_free);
     update_active_session_connection(NULL);
 
-    free(agent_data->screen_info);
-    free(agent_data);
+    g_free(agent_data->screen_info);
+    g_free(agent_data);
+}
+
+static void do_agent_xorg_resolution(struct udscs_connection    **connp,
+                                     struct udscs_message_header *header,
+                                     guint8                      *data)
+{
+    struct agent_data *agent_data = udscs_get_user_data(*connp);
+    guint res_size = sizeof(struct vdagentd_guest_xorg_resolution);
+    guint n = header->size / res_size;
+
+    /* Detect older version session agent, but don't disconnect, as
+     * that stops it from getting the VDAGENTD_VERSION message, and then
+     * it will never re-exec the new version... */
+    if (header->arg1 == 0 && header->arg2 == 0) {
+        syslog(LOG_INFO, "got old session agent xorg resolution message, "
+                         "ignoring");
+        return;
+    }
+
+    if (header->size != n * res_size) {
+        syslog(LOG_ERR, "guest xorg resolution message has wrong size, "
+                        "disconnecting agent");
+        udscs_destroy_connection(connp);
+        return;
+    }
+
+    g_free(agent_data->screen_info);
+    agent_data->screen_info = g_memdup(data, header->size);
+    agent_data->width  = header->arg1;
+    agent_data->height = header->arg2;
+    agent_data->screen_count = n;
+
+    check_xorg_resolution();
+}
+
+static void do_agent_file_xfer_status(struct udscs_connection    **connp,
+                                      struct udscs_message_header *header,
+                                      guint8                      *data)
+{
+    gpointer task_id = GUINT_TO_POINTER(GUINT32_TO_LE(header->arg1));
+    const gchar *log_msg = NULL;
+    guint data_size = 0;
+
+    /* header->arg1 = file xfer task id, header->arg2 = file xfer status */
+    switch (header->arg2) {
+        case VD_AGENT_FILE_XFER_STATUS_NOT_ENOUGH_SPACE:
+            *((guint64 *)data) = GUINT64_TO_LE(*((guint64 *)data));
+            log_msg = "Not enough free space. Cancelling file-xfer %u";
+            data_size = sizeof(guint64);
+            break;
+        case VD_AGENT_FILE_XFER_STATUS_DISABLED:
+            log_msg = "File-xfer is disabled. Cancelling file-xfer %u";
+            break;
+    }
+    send_file_xfer_status(virtio_port, log_msg, header->arg1, header->arg2,
+                          data, data_size);
+
+    if (header->arg2 == VD_AGENT_FILE_XFER_STATUS_CAN_SEND_DATA)
+        g_hash_table_insert(active_xfers, task_id, *connp);
+    else
+        g_hash_table_remove(active_xfers, task_id);
 }
 
 static void agent_read_complete(struct udscs_connection **connp,
     struct udscs_message_header *header, uint8_t *data)
 {
-    struct agent_data *agent_data = udscs_get_user_data(*connp);
-
     switch (header->type) {
-    case VDAGENTD_GUEST_XORG_RESOLUTION: {
-        struct vdagentd_guest_xorg_resolution *res;
-        int n = header->size / sizeof(*res);
-
-        /* Detect older version session agent, but don't disconnect, as
-           that stops it from getting the VDAGENTD_VERSION message, and then
-           it will never re-exec the new version... */
-        if (header->arg1 == 0 && header->arg2 == 0) {
-            syslog(LOG_INFO, "got old session agent xorg resolution message, "
-                             "ignoring");
-            return;
-        }
-
-        if (header->size != n * sizeof(*res)) {
-            syslog(LOG_ERR, "guest xorg resolution message has wrong size, "
-                            "disconnecting agent");
-            udscs_destroy_connection(connp);
-            return;
-        }
-
-        free(agent_data->screen_info);
-        res = malloc(n * sizeof(*res));
-        if (!res) {
-            syslog(LOG_ERR, "out of memory allocating screen info");
-            n = 0;
-        }
-        memcpy(res, data, n * sizeof(*res));
-        agent_data->width  = header->arg1;
-        agent_data->height = header->arg2;
-        agent_data->screen_info  = res;
-        agent_data->screen_count = n;
-
-        check_xorg_resolution();
+    case VDAGENTD_GUEST_XORG_RESOLUTION:
+        do_agent_xorg_resolution(connp, header, data);
         break;
-    }
     case VDAGENTD_CLIPBOARD_GRAB:
     case VDAGENTD_CLIPBOARD_REQUEST:
     case VDAGENTD_CLIPBOARD_DATA:
@@ -926,31 +959,9 @@ static void agent_read_complete(struct udscs_connection **connp,
             return;
         }
         break;
-    case VDAGENTD_FILE_XFER_STATUS:{
-        /* header->arg1 = file xfer task id, header->arg2 = file xfer status */
-        switch (header->arg2) {
-            case VD_AGENT_FILE_XFER_STATUS_NOT_ENOUGH_SPACE: {
-                uint64_t free_space = GUINT64_TO_LE(*((uint64_t*)data));
-                send_file_xfer_status(virtio_port, "Not enough free space. Cancelling file-xfer %u",
-                                      header->arg1, header->arg2,
-                                      (uint8_t*)&free_space, sizeof(uint64_t));
-                break;
-            }
-            case VD_AGENT_FILE_XFER_STATUS_DISABLED:
-                send_file_xfer_status(virtio_port, "File-xfer is disabled, cancelling",
-                                      header->arg1, header->arg2, NULL, 0);
-                break;
-            default:
-                send_file_xfer_status(virtio_port, NULL, header->arg1, header->arg2, NULL, 0);
-        }
-
-        if (header->arg2 == VD_AGENT_FILE_XFER_STATUS_CAN_SEND_DATA)
-            g_hash_table_insert(active_xfers, GUINT_TO_POINTER(GUINT32_TO_LE(header->arg1)),
-                                *connp);
-        else
-            g_hash_table_remove(active_xfers, GUINT_TO_POINTER(GUINT32_TO_LE(header->arg1)));
+    case VDAGENTD_FILE_XFER_STATUS:
+        do_agent_file_xfer_status(connp, header, data);
         break;
-    }
 
     default:
         syslog(LOG_ERR, "unknown message from vdagent: %u, ignoring",
@@ -960,40 +971,29 @@ static void agent_read_complete(struct udscs_connection **connp,
 
 /* main */
 
-static void usage(FILE *fp)
-{
-    fprintf(fp,
-            "Usage: spice-vdagentd [OPTIONS]\n\n"
-            "Spice guest agent daemon, version %s.\n\n"
-            "Options:\n"
-            "  -h             print this text\n"
-            "  -d             log debug messages (use twice for extra info)\n"
-            "  -s <port>      set virtio serial port  [%s]\n"
-            "  -S <filename>  set vdagent Unix domain socket [%s]\n"
-            "  -u <dev>       set uinput device       [%s]\n"
-            "  -f             treat uinput device as fake; no ioctls\n"
-            "  -x             don't daemonize\n"
-            "  -o             only handle one virtio serial session\n"
-#ifdef HAVE_CONSOLE_KIT
-            "  -X             disable console kit integration\n"
-#endif
-#ifdef HAVE_LIBSYSTEMD_LOGIN
-            "  -X             disable systemd-logind integration\n"
-#endif
-            ,VERSION, portdev, vdagentd_socket, uinput_device);
-}
-
 static void daemonize(void)
 {
-    int x;
     FILE *pidfile;
 
     /* detach from terminal */
     switch (fork()) {
     case 0:
-        close(0); close(1); close(2);
+        close(STDIN_FILENO);
+        close(STDOUT_FILENO);
+        close(STDERR_FILENO);
         setsid();
-        x = open("/dev/null", O_RDWR); x = dup(x); x = dup(x);
+        // coverity[leaked_handle] just opening standard file descriptors
+        if (open("/dev/null", O_RDWR) != STDIN_FILENO) {
+            exit(1);
+        }
+        // coverity[leaked_handle] just opening standard file descriptors
+        if (dup(STDIN_FILENO) != STDOUT_FILENO) {
+            exit(1);
+        }
+        // coverity[leaked_handle] just opening standard file descriptors
+        if (dup(STDOUT_FILENO) != STDERR_FILENO) {
+            exit(1);
+        }
         pidfile = fopen(pidfilename, "w");
         if (pidfile) {
             fprintf(pidfile, "%d\n", (int)getpid());
@@ -1003,6 +1003,7 @@ static void daemonize(void)
     case -1:
         syslog(LOG_ERR, "fork: %m");
         retval = 1;
+        // fall through
     default:
         udscs_destroy_server(server);
         exit(retval);
@@ -1081,50 +1082,81 @@ static void quit_handler(int sig)
     quit = 1;
 }
 
+static gboolean parse_debug_level_cb(const gchar *option_name,
+                                     const gchar *value,
+                                     gpointer     data,
+                                     GError     **error)
+{
+    debug++;
+    return TRUE;
+}
+
+static GOptionEntry cmd_entries[] = {
+    { "debug", 'd', G_OPTION_FLAG_NO_ARG,
+      G_OPTION_ARG_CALLBACK, parse_debug_level_cb,
+      "Log debug messages (use twice for extra info)", NULL },
+
+    { "virtio-serial-port-path", 's', 0,
+      G_OPTION_ARG_STRING, &portdev,
+      "Set virtio-serial path (" DEFAULT_VIRTIO_PORT_PATH ")", NULL },
+
+    { "vdagentd-socket", 'S', 0,
+      G_OPTION_ARG_STRING, &vdagentd_socket,
+       "Set spice-vdagentd socket (" VDAGENTD_SOCKET ")", NULL },
+
+    { "uinput-device", 'u', 0,
+      G_OPTION_ARG_STRING, &uinput_device,
+      "Set uinput device (" DEFAULT_UINPUT_DEVICE ")", NULL },
+
+    { "fake-uinput", 'f', 0,
+      G_OPTION_ARG_NONE, &uinput_fake,
+      "Treat uinput device as fake; no ioctls", NULL },
+
+    { "foreground", 'x', G_OPTION_FLAG_REVERSE,
+      G_OPTION_ARG_NONE, &do_daemonize,
+      "Do not daemonize the agent", NULL},
+
+    { "one-session", 'o', 0,
+      G_OPTION_ARG_NONE, &only_once,
+      "Only handle one virtio serial session", NULL },
+
+#if defined(HAVE_CONSOLE_KIT) || defined (HAVE_LIBSYSTEMD_LOGIN)
+    { "disable-session-integration", 'X', G_OPTION_FLAG_REVERSE,
+      G_OPTION_ARG_NONE, &want_session_info,
+      "Disable console kit and systemd-logind integration", NULL },
+#endif
+
+    { NULL }
+};
+
 int main(int argc, char *argv[])
 {
-    int c;
-    int do_daemonize = 1;
-    int want_session_info = 1;
+    GOptionContext *context;
+    GError *err = NULL;
     struct sigaction act;
     gboolean own_socket = TRUE;
 
-    for (;;) {
-        if (-1 == (c = getopt(argc, argv, "-dhxXfos:u:S:")))
-            break;
-        switch (c) {
-        case 'd':
-            debug++;
-            break;
-        case 's':
-            portdev = optarg;
-            break;
-        case 'S':
-            vdagentd_socket = optarg;
-            break;
-        case 'u':
-            uinput_device = optarg;
-            break;
-        case 'f':
-            uinput_fake = 1;
-            break;
-        case 'o':
-            only_once = 1;
-            break;
-        case 'x':
-            do_daemonize = 0;
-            break;
-        case 'X':
-            want_session_info = 0;
-            break;
-        case 'h':
-            usage(stdout);
-            return 0;
-        default:
-            fputs("\n", stderr);
-            usage(stderr);
-            return 1;
-        }
+    context = g_option_context_new(NULL);
+    g_option_context_add_main_entries(context, cmd_entries, NULL);
+    g_option_context_set_summary(context,
+        "Spice guest agent daemon, version " VERSION);
+    g_option_context_parse(context, &argc, &argv, &err);
+    g_option_context_free(context);
+
+    if (err) {
+        g_printerr("Invalid arguments, %s\n", err->message);
+        g_error_free(err);
+        return 1;
+    }
+
+    if (portdev == NULL) {
+        portdev = g_strdup(DEFAULT_VIRTIO_PORT_PATH);
+    }
+    if (vdagentd_socket == NULL) {
+        vdagentd_socket = g_strdup(VDAGENTD_SOCKET);
+    }
+    if (uinput_device == NULL) {
+        uinput_device = g_strdup(DEFAULT_UINPUT_DEVICE);
     }
 
     memset(&act, 0, sizeof(act));
@@ -1149,8 +1181,7 @@ int main(int argc, char *argv[])
         server = udscs_create_server_for_fd(SD_LISTEN_FDS_START, agent_connect,
                                             agent_read_complete,
                                             agent_disconnect,
-                                            vdagentd_messages,
-                                            VDAGENTD_NO_MESSAGES, debug);
+                                            debug);
         own_socket = FALSE;
     } else
     /* systemd socket activation not enabled, create our own */
@@ -1158,7 +1189,6 @@ int main(int argc, char *argv[])
     {
         server = udscs_create_server(vdagentd_socket, agent_connect,
                                      agent_read_complete, agent_disconnect,
-                                     vdagentd_messages, VDAGENTD_NO_MESSAGES,
                                      debug);
     }
 
@@ -1166,8 +1196,6 @@ int main(int argc, char *argv[])
         if (errno == EADDRINUSE) {
             syslog(LOG_CRIT, "Fatal the server socket %s exists already. Delete it?",
                    vdagentd_socket);
-        } else if (errno == ENOMEM) {
-            syslog(LOG_CRIT, "Fatal could not allocate memory for udscs server");
         } else {
             syslog(LOG_CRIT, "Fatal could not create the server socket %s: %m",
                    vdagentd_socket);
@@ -1223,5 +1251,9 @@ int main(int argc, char *argv[])
     if (do_daemonize)
         unlink(pidfilename);
 
+    g_free(portdev);
+    g_free(vdagentd_socket);
+    g_free(uinput_device);
+
     return retval;
 }
